/*
 * COPYRIGHT (c) International Business Machines Corp. 2021
 *
 * This program is provided under the terms of the Common Public License,
 * version 1.0 (CPL-1.0). Any use, reproduction or distribution for this
 * software constitutes recipient's acceptance of CPL-1.0 terms which can be
 * found in the file LICENSE file or at
 * https://opensource.org/licenses/cpl1.0.php
 */
#ifndef UNITTEST_H
#define UNITTEST_H

/* AUTOTOOLS definitions for tests */
#define TEST_SKIP 77
#define TEST_FAIL 99
#define TEST_PASS  0

#define ARRAYSIZE(X) (sizeof((X))/sizeof((X)[0]))

#endif
